@@ -1,6 +1,6 @@
-;;; geiser-syntax.el -- utilities for parsing scheme syntax
+;;; geiser-syntax.el --- Utilities for parsing scheme syntax  -*- lexical-binding: t; -*-
 
-;; Copyright (C) 2009, 2010, 2011, 2012, 2013, 2014, 2015 Jose Antonio Ortega Ruiz
+;; Copyright (C) 2009-2016, 2019-2022 Jose Antonio Ortega Ruiz
 
 ;; This program is free software; you can redistribute it and/or
 ;; modify it under the terms of the Modified BSD License. You should
@@ -10,6 +10,7 @@
 ;; Start date: Sun Feb 08, 2009 15:03
 
 
+;;; Code:
 
 (require 'geiser-impl)
 (require 'geiser-popup)
@@ -17,7 +18,9 @@
 
 (require 'scheme)
 
-(eval-when-compile (require 'cl))
+(eval-when-compile
+  (require 'cl-lib)
+  (require 'subr-x))
 
 
 ;;; Indentation:
@@ -41,7 +44,7 @@
  (match 1)
  (match-lambda 0)
  (match-lambda* 0)
- (match-let 1)
+ (match-let scheme-let-indent)
  (match-let* 1)
  (match-letrec 1)
  (opt-lambda 1)
@@ -55,6 +58,7 @@
  (test-eq 1)
  (test-equal 1)
  (test-eqv 1)
+ (test-group 1)
  (test-group-with-cleanup 1)
  (test-runner-on-bad-count! 1)
  (test-runner-on-bad-end-name! 1)
@@ -134,13 +138,20 @@ implementation-specific entries for font-lock-keywords.")
 (geiser-impl--define-caller geiser-syntax--case-sensitive case-sensitive ()
   "A flag saying whether keywords are case sensitive.")
 
-(defun geiser-syntax--add-kws ()
-  (when (not (and (boundp 'quack-mode) quack-mode))
+(defun geiser-syntax--add-kws (&optional global-p)
+  (unless (bound-and-true-p quack-mode)
     (let ((kw (geiser-syntax--impl-kws geiser-impl--implementation))
           (cs (geiser-syntax--case-sensitive geiser-impl--implementation)))
       (when kw (font-lock-add-keywords nil kw))
+      (when global-p (font-lock-add-keywords nil (geiser-syntax--keywords)))
       (setq font-lock-keywords-case-fold-search (not cs)))))
 
+(defun geiser-syntax--remove-kws ()
+  (unless (bound-and-true-p quack-mode)
+    (let ((kw (geiser-syntax--impl-kws geiser-impl--implementation)))
+      (when kw
+        (font-lock-remove-keywords nil kw)))))
+
 
 ;;; A simple scheme reader
 
@@ -181,6 +192,13 @@ implementation-specific entries for font-lock-keywords.")
   (geiser-syntax--read/token
    (cons 'unprintable (geiser-syntax--read/matching "<" ">"))))
 
+(defun geiser-syntax--read/ex-symbol ()  ;; #{foo bar}# style symbols
+  (let ((tk (geiser-syntax--read/matching "{" "}")))
+    (when-let (c (geiser-syntax--read/next-char))
+      (when (char-equal ?\# c)
+        (geiser-syntax--read/next-char)
+        (cons 'atom (make-symbol (format "#{%s}#" tk)))))))
+
 (defun geiser-syntax--read/skip-comment ()
   (while (and (geiser-syntax--read/next-char)
               (nth 8 (syntax-ppss))))
@@ -189,15 +207,15 @@ implementation-specific entries for font-lock-keywords.")
 (defun geiser-syntax--read/next-token ()
   (skip-syntax-forward "->")
   (if (geiser-syntax--read/eos) '(eob)
-    (case (char-after)
+    (cl-case (char-after)
       (?\; (geiser-syntax--read/skip-comment))
       ((?\( ?\[) (geiser-syntax--read/token 'lparen))
       ((?\) ?\]) (geiser-syntax--read/token 'rparen))
       (?. (if (memq (car (syntax-after (1+ (point)))) '(0 11 12))
               (geiser-syntax--read/token 'dot)
             (cons 'atom (geiser-syntax--read/elisp))))
-      (?\# (case (geiser-syntax--read/next-char)
-             ('nil '(eob))
+      (?\# (cl-case (geiser-syntax--read/next-char)
+             ((nil quote) '(eob))
              (?| (geiser-syntax--read/skip-comment))
              (?: (if (geiser-syntax--read/next-char)
                      (cons 'kwd (geiser-syntax--read/symbol))
@@ -206,11 +224,20 @@ implementation-specific entries for font-lock-keywords.")
              (?\( (geiser-syntax--read/token 'vectorb))
              (?\< (geiser-syntax--read/unprintable))
              ((?' ?` ?,) (geiser-syntax--read/next-token))
+             (?\{ (geiser-syntax--read/ex-symbol))
              (t (let ((tok (geiser-syntax--read/symbol)))
                   (cond ((equal (symbol-name tok) "t") '(boolean . :t))
                         ((equal (symbol-name tok) "f") '(boolean . :f))
                         (tok (cons 'atom tok))
                         (t (geiser-syntax--read/next-token)))))))
+      (?| (cl-case (geiser-syntax--read/next-char) ;; gambit style block comments
+            ((nil quote) '(eob))
+            (?# (geiser-syntax--read/skip-comment))
+            (t (let ((tok (geiser-syntax--read/symbol)))
+                 (cond ((equal (symbol-name tok) "t") '(boolean . :t))
+                       ((equal (symbol-name tok) "f") '(boolean . :f))
+                       (tok (cons 'atom tok))
+                       (t (geiser-syntax--read/next-token)))))))
       (?\' (geiser-syntax--read/token '(quote . quote)))
       (?\` (geiser-syntax--read/token
             `(backquote . ,backquote-backquote-symbol)))
@@ -219,7 +246,8 @@ implementation-specific entries for font-lock-keywords.")
                `(splice . ,backquote-splice-symbol))
             `(unquote . ,backquote-unquote-symbol)))
       (?\" (cons 'string (geiser-syntax--read/elisp)))
-      (t (cons 'atom (geiser-syntax--read/symbol))))))
+      (t (let ((x (geiser-syntax--read/elisp)))
+           (cons 'atom (if (atom x) x (geiser-syntax--read/symbol))))))))
 
 (defsubst geiser-syntax--read/match (&rest tks)
   (let ((token (geiser-syntax--read/next-token)))
@@ -251,7 +279,7 @@ implementation-specific entries for font-lock-keywords.")
 (defun geiser-syntax--read ()
   (let ((token (geiser-syntax--read/next-token))
         (max-lisp-eval-depth (max max-lisp-eval-depth 3000)))
-    (case (car token)
+    (cl-case (car token)
       (eob nil)
       (lparen (geiser-syntax--read/list))
       (vectorb (apply 'vector (geiser-syntax--read/list)))
@@ -265,20 +293,24 @@ implementation-specific entries for font-lock-keywords.")
 
 (defun geiser-syntax--read-from-string (string &optional start end)
   (when (stringp string)
-    (let* ((start (or start 0))
-           (end (or end (length string)))
-           (max-lisp-eval-depth (min 20000
-                                     (max max-lisp-eval-depth (- end start)))))
-      (with-temp-buffer
-        (save-excursion (insert string))
-        (cons (ignore-errors (geiser-syntax--read)) (point))))))
+    ;; In Emacs 29 this variable doesn't have an effect
+    ;; anymore and `max-lisp-eval-depth' achieves the same.
+    (with-suppressed-warnings ((obsolete max-specpdl-size))
+      (let* ((start (or start 0))
+             (end (or end (length string)))
+             (max-lisp-eval-depth (min 20000
+                                       (max max-lisp-eval-depth (- end start))))
+             (max-specpdl-size (* 2 max-lisp-eval-depth)))
+        (with-temp-buffer
+          (save-excursion (insert string))
+          (cons (geiser-syntax--read) (point)))))))
 
 (defun geiser-syntax--form-from-string (s)
   (car (geiser-syntax--read-from-string s)))
 
 (defsubst geiser-syntax--form-after-point (&optional boundary)
   (let ((geiser-syntax--read/buffer-limit (and (numberp boundary) boundary)))
-    (save-excursion (values (geiser-syntax--read) (point)))))
+    (save-excursion (list (geiser-syntax--read) (point)))))
 
 (defun geiser-syntax--mapconcat (fun lst sep)
   (cond ((null lst) "")
@@ -335,7 +367,7 @@ implementation-specific entries for font-lock-keywords.")
 (defsubst geiser-syntax--symbol-eq (s0 s1)
   (and (symbolp s0) (symbolp s1) (equal (symbol-name s0) (symbol-name s1))))
 
-(defun geiser-syntax--scan-sexps (&optional begin)
+(defun geiser-syntax--scan-sexps ()
   (let* ((fst (geiser-syntax--symbol-at-point))
          (smth (or fst (not (looking-at-p "[\s \s)\s>\s<\n]"))))
          (path (and fst `((,fst 0)))))
@@ -383,30 +415,31 @@ implementation-specific entries for font-lock-keywords.")
 
 (defun geiser-syntax--linearize (form)
   (cond ((not (listp form)) (list form))
-	((null form) nil)
-	(t (cons (car form) (geiser-syntax--linearize (cdr form))))))
+        ((null form) nil)
+        (t (cons (car form) (geiser-syntax--linearize (cdr form))))))
 
 (defun geiser-syntax--scan-locals (bfs sbfs form nesting locals)
   (if (or (null form) (not (listp form)))
       (geiser-syntax--normalize locals)
     (if (not (geiser-syntax--binding-form-p bfs sbfs (car form)))
-	(geiser-syntax--scan-locals bfs sbfs
-				    (car (last form))
+        (geiser-syntax--scan-locals bfs sbfs
+                                    (car (last form))
                                     (1- nesting) locals)
       (let* ((head (car form))
-	     (name (geiser-syntax--if-symbol (cadr form)))
-	     (names (if name (geiser-syntax--if-list (caddr form))
-		      (geiser-syntax--if-list (cadr form))))
+             (name (geiser-syntax--if-symbol (cadr form)))
+             (names (if name (geiser-syntax--if-list (caddr form))
+                      (geiser-syntax--if-list (cadr form))))
              (bns (and name
                        (geiser-syntax--binding-form-p bfs sbfs (car names))))
-	     (rest (if (and name (not bns)) (cdddr form) (cddr form)))
-	     (use-names (and (or rest
+             (rest (if (and name (not bns)) (cdddr form) (cddr form)))
+             (use-names (and (or rest
                                  (< nesting 1)
                                  (geiser-syntax--binding-form*-p sbfs head))
                              (not bns))))
-	(when name (push name locals))
+        (when name (push name locals))
         (when (geiser-syntax--symbol-eq head 'case-lambda)
-          (dolist (n (and (> nesting 0) (caar (last form))))
+          (dolist (n (and (> nesting 0)
+                          (geiser-syntax--linearize (caar (last form)))))
             (when n (push n locals)))
           (setq rest (and (> nesting 0) (cdr form)))
           (setq use-names nil))
@@ -414,19 +447,19 @@ implementation-specific entries for font-lock-keywords.")
           (dolist (n (and (> nesting 0) (cdaar (last form))))
             (when n (push n locals)))
           (setq rest (and (> nesting 0) (cdr form))))
-	(when use-names
-	  (dolist (n (geiser-syntax--linearize names))
+        (when use-names
+          (dolist (n (geiser-syntax--linearize names))
             (let ((xs (if (and (listp n) (listp (car n))) (car n) (list n))))
               (dolist (x xs) (when x (push x locals))))))
-	(dolist (f (butlast rest))
-	  (when (and (listp f)
+        (dolist (f (butlast rest))
+          (when (and (listp f)
                      (geiser-syntax--symbol-eq (car f) 'define)
                      (cadr f))
-	    (push (cadr f) locals)))
-	(geiser-syntax--scan-locals bfs sbfs
-				    (car (last (or rest names)))
+            (push (cadr f) locals)))
+        (geiser-syntax--scan-locals bfs sbfs
+                                    (car (last (or rest names)))
                                     (1- nesting)
-				    locals)))))
+                                    locals)))))
 
 (defun geiser-syntax--locals-around-point (bfs sbfs)
   (when (eq major-mode 'scheme-mode)
@@ -437,7 +470,7 @@ implementation-specific entries for font-lock-keywords.")
         (let ((boundary (point))
               (nesting (geiser-syntax--nesting-level)))
           (geiser-syntax--pop-to-top)
-          (multiple-value-bind (form end)
+          (cl-destructuring-bind (form _end)
               (geiser-syntax--form-after-point boundary)
             (delete sym
                     (geiser-syntax--scan-locals bfs
@@ -465,23 +498,71 @@ implementation-specific entries for font-lock-keywords.")
                  (geiser-syntax--mapconcat 'geiser-syntax--display a " ")))
         (t (format "%s" a))))
 
+(defconst geiser-syntax--font-lock-buffer-name " *Geiser font-lock*")
+
+(defun geiser-syntax--font-lock-buffer-p (&optional buffer)
+  (equal (buffer-name buffer) geiser-syntax--font-lock-buffer-name))
+
 (defun geiser-syntax--font-lock-buffer ()
-  (let ((name " *geiser font lock*"))
-    (or (get-buffer name)
-        (let ((buffer (get-buffer-create name)))
-          (set-buffer buffer)
-          (let ((geiser-default-implementation
-                 (or geiser-default-implementation
-                     (car geiser-active-implementations))))
-            (scheme-mode))
-          buffer))))
+  (or (get-buffer geiser-syntax--font-lock-buffer-name)
+      (let ((buffer (get-buffer-create geiser-syntax--font-lock-buffer-name)))
+        (set-buffer buffer)
+        (let ((geiser-default-implementation
+               (or geiser-default-implementation
+                   (car geiser-active-implementations))))
+          (scheme-mode))
+        buffer)))
+
+(defun geiser-syntax--fontify (&optional beg end)
+  (let ((font-lock-verbose nil)
+        (beg (or beg (point-min)))
+        (end (or end (point-max))))
+    (if (fboundp 'font-lock-flush)
+        (font-lock-flush beg end)
+      (with-no-warnings (font-lock-fontify-region beg end)))))
+
+;; derived from org-src-font-lock-fontify-block (org-src.el)
+(defun geiser-syntax--fontify-syntax-region (start end)
+  "Fontify region as Scheme."
+  (let ((string (buffer-substring-no-properties start end))
+        (modified (buffer-modified-p))
+        (buffer-undo-list t)
+        (geiser-buffer (current-buffer)))
+    (with-current-buffer
+        (get-buffer-create " *Geiser REPL fontification*")
+      (let ((inhibit-modification-hooks nil))
+        (erase-buffer)
+        ;; Add string and a final space to ensure property change.
+        (insert string " "))
+      ;; prevent geiser prompt
+      (let ((geiser-default-implementation
+             (or geiser-default-implementation
+                 (car geiser-active-implementations))))
+        (scheme-mode))
+      (geiser--font-lock-ensure)
+      (let ((pos (point-min)) next)
+        (while (setq next (next-property-change pos))
+          ;; Handle additional properties from font-lock, so as to
+          ;; preserve, e.g., composition.
+          (dolist (prop (cons 'face font-lock-extra-managed-props))
+            (let ((new-prop (get-text-property pos prop))
+                  (start-point (+ start (1- pos)))
+                  (end-point (1- (+ start next))))
+              (put-text-property start-point end-point prop new-prop geiser-buffer)))
+          (setq pos next))))
+    (add-text-properties
+     start end
+     '(font-lock-fontified t
+                           fontified t
+                           font-lock-multiline t))
+    (set-buffer-modified-p modified)))
 
 (defun geiser-syntax--scheme-str (str)
   (save-current-buffer
     (set-buffer (geiser-syntax--font-lock-buffer))
     (erase-buffer)
     (insert str)
-    (let ((font-lock-verbose nil)) (font-lock-ensure))
+    (geiser-syntax--fontify)
     (buffer-string)))
 
 
